from ..decorators import *
from typing import *
from . import Plugin


class CalculatorPlugin(Plugin):
    @function
    def evaluate(
        self,
        expression: str = param(
            "The math expression to evaluate. Must be an valid python expression."
        ),
    ):
        """Execute a math expression and return the result. The expression must be an valid python expression that can be execuated by `eval()`."""
        self._log_call("evaluate", expression)
        result = eval(expression)
        return result



from ..decorators import *
from typing import *
from ..message import Message
import openai
import json
import pandas as pd
from . import Plugin
import os
from openai.embeddings_utils import get_embedding, cosine_similarity
import numpy as np
import datetime


class MemoryPlugin(Plugin):
    def __init__(self, data_file: str = "_memory-with-embedding.csv"):
        self.data_file = data_file

    def __load_data(self):
        return (
            pd.read_csv(self.data_file)
            if os.path.isfile(self.data_file) and os.path.getsize(self.data_file) > 0
            else pd.DataFrame(columns=["timestamp", "text", "embedding"])
        )

    def __save_new_data(self, text: str):
        df = self.__load_data()
        embedding = self.__get_embedding(text, model="text-embedding-ada-002")
        df.loc[len(df.index)] = [datetime.datetime.now().isoformat(), text, embedding]
        df.to_csv(self.data_file, index=False)

    def __get_embedding(self, text, model="text-embedding-ada-002"):
        # text = text.replace("\n", " ")
        return openai.Embedding.create(input=[text], model=model)["data"][0][
            "embedding"
        ]

    def clear_memory(self):
        f = open(self.data_file, "w")
        f.close()

    @function
    def search_from_memory(
        self,
        query: str = param("The query of the search"),
    ):
        """Search content from the long term memory, by using the given query. The memory includes all the prior chat history and prior knowledge.
        The search results are in descending order in importance. Always respect to the top results.
        If you're unsure about some past events or some knowledge, always check the memory first as you may forgot it.
        """
        self._log_call("search_from_memory", query)
        embedding = self.__get_embedding(query)
        df = self.__load_data()
        df["embedding"] = df.embedding.apply(eval).apply(np.array)
        df["similarity"] = df.embedding.apply(lambda x: cosine_similarity(x, embedding))
        data = df.sort_values("similarity", ascending=False)
        results = ""
        n = 0
        for index, row in data.iterrows():
            results += "TIMESTAMP: " + row["timestamp"] + "\n"
            results += "CONTENT:\n" + row["text"] + "\n\n---\n\n"
            if n >= 2:
                break
            n += 1
        return results

    @function
    def remember(
        self,
        content: str = param("The content to remember in the memory"),
    ):
        """Actively remember some information as part of the long-term memory, if it is important."""
        self._log_call("remember", content)
        self.__save_new_data(content)
        return "done."

    def on_new_chat_message(self, msg: Message):
        msg_s = json.dumps(msg.to_json())
        self.__save_new_data(msg_s)



from typing import *
from .message import *
from dotenv import dotenv_values
import inspect
from inspect import Parameter
import openai
import logging
import os

openai_api_key = None


class AugmentedGPT:
    @staticmethod
    def set_api_key(key: str):
        global openai_api_key
        openai_api_key = key

    def __init__(
        self,
        functions: List[Callable] = [],
        plugins: List["Plugin"] = [],
        debug=False,
    ):
        api_key = dotenv_values().get(
            "OPENAI_API_KEY", os.environ.get("OPENAI_API_KEY", openai_api_key)
        )
        assert api_key is not None, "Missing OPENAI_API_KEY"
        openai.api_key = api_key
        self.logger = logging.getLogger("AugmentedGPT")
        self.logger.setLevel(logging.DEBUG if debug else logging.INFO)
        self.__functions: Dict[str, Tuple[Any, Callable]] = {}
        for f in functions:
            self.add_function(f)
        self.__plugins = {}
        for p in plugins:
            clsname = p.__class__.__name__
            if clsname.endswith("Plugin"):
                clsname = clsname[:-6]
            p.register(self)
            self.__plugins[clsname] = p
        self.history: List[Message] = []

    def get_plugin(self, name: str) -> "Plugin":
        return self.__plugins[name]

    def add_function(self, f: Callable):
        func_info = getattr(f, "gpt_function_call_info")
        self.logger.info("Register-Function: " + func_info["name"])
        self.__functions[func_info["name"]] = (func_info, f)

    def __filter_args(self, callable: Callable, args: Any):
        p_args = []
        kw_args = {}
        for p in inspect.signature(callable).parameters.values():
            match p.kind:
                case Parameter.POSITIONAL_ONLY:
                    p_args.append(args[p.name] if p.name in args else p.default.default)
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY:
                    kw_args[p.name] = (
                        args[p.name] if p.name in args else p.default.default
                    )
                case other:
                    raise ValueError(f"{other} is not supported")
        return p_args, kw_args

    def __call_function(self, function_call) -> Message:
        func_name = function_call["name"]
        func = self.__functions[func_name][1]
        arguments = json.loads(function_call["arguments"])
        args, kw_args = self.__filter_args(func, arguments)
        result = func(*args, **kw_args)
        if not isinstance(result, str):
            result = json.dumps(result)
        return Message(role="function", name=func_name, content=result)

    def __chat_completion_request(
        self, messages: List[Message], stream=False
    ) -> Union[Message, MessageStream]:
        functions = [x for (x, _) in self.__functions.values()]
        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo-0613",
            messages=[m.to_json() for m in messages],
            functions=functions,
            function_call="auto",
            stream=stream,
        )
        if stream:
            return MessageStream(response)
        else:
            return Message(**response["choices"][0]["message"])

    def chat_completion(
        self,
        messages: Union[Message, List[Message]],
        stream=False,
        content_free=False,
    ):
        history = self.history if not content_free else []
        if isinstance(messages, list):
            history.extend(messages)
            for m in messages:
                self.__on_new_chat_message(m)
        else:
            history.append(messages)
            self.__on_new_chat_message(messages)
        response = self.__chat_completion_request(history, stream=stream)
        yield response
        message = response.message() if stream else response
        history.append(message)
        self.__on_new_chat_message(message)
        while message.function_call is not None:
            result = self.__call_function(message.function_call)
            history.append(result)
            self.__on_new_chat_message(result)
            yield MessageStream(None, result) if stream else result
            response = self.__chat_completion_request(history, stream=stream)
            yield response
            message = response.message() if stream else response
            history.append(message)
            self.__on_new_chat_message(message)

    def __on_new_chat_message(self, msg: Message):
        for p in self.__plugins.values():
            p.on_new_chat_message(msg)



import logging

logging.basicConfig(format="[%(levelname)s:%(name)s] %(message)s")

from .decorators import *
from .message import *
from .plugins import Plugin
from . import plugins
from .augmented_gpt import AugmentedGPT

__all__ = [
    "param",
    "function",
    "Message",
    "MessageStream",
    "Plugin",
    "plugins",
    "AugmentedGPT",
]



import inspect
from ..decorators import *
from ..message import Message
import datetime
from logging import Logger


class Plugin:
    logger: Logger

    def register(self, gpt: "AugmentedGPT"):
        for n, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not hasattr(method, "gpt_function_call_info"):
                continue
            func_info = getattr(method, "gpt_function_call_info")
            clsname = self.__class__.__name__
            if clsname.endswith("Plugin"):
                clsname = clsname[:-6]
            if not func_info["name"].startswith(clsname + "-"):
                func_info["name"] = clsname + "-" + func_info["name"]
            gpt.add_function(method)
        self.logger = gpt.logger

    def _log_call(self, name: str, *args, **kwargs):
        msg = f"➡️ {name}: " + ", ".join(args) + ", ".join((f"{k}={v}" for k, v in kwargs.items()))
        self.logger.debug(msg)

    def on_new_chat_message(self, msg: Message):
        ...


class TimestampPlugin(Plugin):
    @function
    def get_current_timestamp(self):
        """Get the current tempstamp in ISO format"""
        return datetime.datetime.now().isoformat()

from .memory import *
from .calc import *


def all_plugins():
    return [TimestampPlugin(), MemoryPlugin(), CalculatorPlugin()]



from typing import *
import json
from dataclasses import dataclass

JSON: TypeAlias = dict[str, "JSON"] | list["JSON"] | str | int | float | bool | None


@dataclass
class Message:
    role: Literal["system", "user", "assistant", "function"]
    content: str
    name: Optional[str] = None
    function_call: Optional[JSON] = None

    def __post_init__(self):
        if self.function_call is not None:
            self.function_call = json.loads(json.dumps((self.function_call)))

    def to_json(self):
        data = {
            "role": self.role,
            "content": self.content,
        }
        if self.name is not None:
            data["name"] = self.name
        if self.function_call is not None:
            data["function_call"] = self.function_call
        return data

    def message(self) -> "Message":
        return self


class MessageStream:
    def __init__(self, response, final_message=None):
        self.__response = response
        self.__message = {}
        self.__final_message = final_message

    def __next__(self):
        if self.__final_message is not None:
            raise StopIteration()
        chunk = next(self.__response)
        delta = chunk["choices"][0]["delta"]

        def concat(target, delta):
            for k, v in delta.items():
                if isinstance(v, str) or v is None:
                    if k not in target:
                        target[k] = ""
                    target[k] += v or ""
                else:
                    if k not in target:
                        target[k] = {}
                    concat(target[k], v)

        concat(self.__message, delta)
        return delta

    def __iter__(self):
        return self

    def message(self) -> Message:
        if self.__final_message is not None:
            return self.__final_message
        for _ in self:
            ...
        return Message(**self.__message)



from typing import *
import inspect
from inspect import Parameter


class _Param:
    def __init__(
        self,
        description: Optional[str] = None,
        default: Optional[Any] = None,
        enum: Optional[List[Any]] = None,
    ):
        self.description = description
        self.default = default
        self.enum = enum


def param(
    description: Optional[str] = None,
    default: Optional[Any] = None,
    enum: Optional[List[Any]] = None,
) -> Any:
    return _Param(description, default, enum)


def function(
    description: Optional[str] = None,
    name: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
):
    def decorator_func(callable: Callable):
        # Get function name
        fname = name if name is not None else callable.__name__
        # Get parameter info
        params = {
            "type": "object",
            "properties": {},
            "required": [],
        }
        for pname, param in inspect.signature(callable).parameters.items():
            if pname == "self":
                continue
            assert isinstance(
                param.default, _Param
            ), f"Invalid default value for parameter `{pname}` in function {fname}"
            prop = {}
            required = True
            match param.annotation:
                case x if x == str or x == Optional[str]:
                    required = x == str
                    prop["type"] = "string"
                case x if x == int or x == Optional[int]:
                    required = x == int
                    prop["type"] = "integer"
                case other:
                    assert (
                        False
                    ), f"Invalid type annotation for parameter `{pname}` in function {fname}"
            if required:
                params["required"].append(pname)
            else:
                assert (
                    param.default.default is not None
                ), f"Optional parameter `{pname}` in function {fname} requires a default value"
            params["properties"][pname] = prop
        setattr(
            callable,
            "gpt_function_call_info",
            {
                "name": fname,
                "description": description or callable.__doc__ or "",
                "parameters": params,
            },
        )
        return callable

    if inspect.isfunction(description) or inspect.ismethod(description):
        func = description
        description = None
        return decorator_func(func)
    return decorator_func
